import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;

//----------------------------------------------------------------------

/**
 * Main class.
 */
public class Main {

	public static final int HEIGHT = 768;
	public static final int WIDTH  = 1024;
	public static final Color BACKGROUND_COLOR = Color.white;
	
	public static final int LINEHEIGHT = 10;
	public static final int MARGIN = 30;

	public static MainFrame root;
	public static Score     score;
	public static Toolkit   toolkit = Toolkit.getDefaultToolkit();
	public static Undo      undo;

	public static final Image noteheadImage  = toolkit.getImage("image/notehead.png");
	public static final Image noteheadImage2 = toolkit.getImage("image/notehead2.png");

	static BasicStroke doubledStroke = new BasicStroke(2.0f);
	static BasicStroke fifthtimesStroke = new BasicStroke(5.0f);

	public static void main(String args[]) {
		root = new MainFrame("MainFrame");
		score = new Score();
		undo = new Undo(root.buttonPanel.undoButton);
		root.setVisible(true);
	}

	// ONLY FOR CONVENIENCE
	// debug Ѥ print
	public static void debugPrint(int level, String str) {
		if (level > 2) {
			System.out.println(str);
		}
	}

	// ONLY FOR CONVENIENCE
	// what shouldn't happen Ȥ˻Ȥ
	public static void failwith(String str) {
		System.out.println("failwith :" + str);
		System.exit(1);
	}

	// ONLY FOR CONVENIENCE
	// mainPanel  graphics ֥Ȥ롣
	public static Graphics getGraphics() {
		return root.mainPanel.getImage().getGraphics();
	}

}

//----------------------------------------------------------------------

/**
 * MainFrame
 * ޤΤȤButtonPanel  MainPanel Ĥġ
 */
class MainFrame extends Frame {

	ButtonPanel buttonPanel;
	MainPanel mainPanel;

	MainFrame(String str) {
		super(str);
		setSize(Main.WIDTH, Main.HEIGHT);
		setBackground(Main.BACKGROUND_COLOR);
		setVisible(true);

		addWindowListener(new WindowAdapter() {
				public void windowClosing(WindowEvent we) {
					System.exit(0);
				}});

		mainPanel = new MainPanel();
		buttonPanel = new ButtonPanel();

		// դ褯ʬޤ󤬡
		// ޤȤľǤҤȤġ
		// ʤ swing 礭Ѥ褦Ȥȡ۾˽ŤʤǤ
		// ȡ ScrollBar Ȥäơ岼Υ뤬褦ˤ٤
		setLayout(null);
		add(buttonPanel);
		add(mainPanel);
		buttonPanel.setBounds(0, 60, 1024, 60);
		mainPanel.setBounds(0, 120, 1024, 700);

		this.show();
	}

}

//----------------------------------------------------------------------------------------------------

/**
 *  Button ¤٤뤿ˤĤѥͥ롣
 *  Button ϡinit, repaint ȤꤢѰաȤɬפ˱ɲä롣
 */
class ButtonPanel extends Panel implements ActionListener {

	Button initButton    = new Button("Init");
	Button undoButton    = new Button("Undo");
	Button playButton    = new Button("Play");
	Button repaintButton = new Button("Repaint");

	ButtonPanel() {

		initButton.addActionListener(this);
		undoButton.addActionListener(this);
		playButton.addActionListener(this);
		repaintButton.addActionListener(this);

		undoButton.setEnabled(false);
									 
		add(initButton);
		add(undoButton);
		add(playButton);
		add(repaintButton);
		
	}

	public void actionPerformed(ActionEvent ae) {
		Object src = ae.getSource();

		if (src == initButton) {
			Main.score.init();
			Main.root.mainPanel.repaint();
			Main.undo.clear();
			Main.debugPrint(1, "initButton pressed");
		} else if (src == undoButton) {
			if (Main.undo.canUndo()) {
				Main.undo.undo();
			}
			Main.root.mainPanel.repaint();
		} else if (src == playButton) {
			LinkedList[] notes = Main.score.getNotes();
			Thread[] threads = new Thread[notes.length];
			for (int i = 0; i < notes.length; ++i) {
				Play play = new Play(120, notes[i]);
				threads[i] = new Thread(play);
			}

			for (int i = 0; i < notes.length; ++i) {
				threads[i].start();
			}
		} else if (src == repaintButton) {
			Main.root.mainPanel.repaint();
		} else {
			Main.failwith("actionPerformed");
		}
	}

}
//----------------------------------------------------------------------------------------------------

/**
 *
 */
class MainPanel extends Panel {

	Image image = null;

	Point prevPoint = new Point(-1, -1);
	LinkedList points = null;

	MainPanel() {
		super();
		MouseDispatcher mouseDispatcher = new MouseDispatcher(this);
		addMouseListener(mouseDispatcher);
		addMouseMotionListener(mouseDispatcher);

		this.setVisible(true);
	}

	// ºݤβ̾ΰ ( mainPanel а) 顢
	// mainPanel а֤׻֤
	// ʤ뤬ȯƤʤɤ˰򤹤롣
	// ޤ⤷Ƥʤ
	Point getInternalPoint(Point p) {
		return p;
	}

	public Image getImage() {
		Dimension d = getSize();
		if (image == null 
			|| image.getHeight(this) != d.height 
			|| image.getWidth(this) != d.width) {
			image = createImage(d.width, d.height);
		}
		return image;
	}

	public void update(Graphics g) {
		paint(g);
	}

	public void paint(Graphics g) {
		Dimension d = getSize();
		getImage().getGraphics().clearRect(0, 0, d.width, d.height);
		Main.score.paintStaffNotation();
		g.drawImage(getImage(), 0, 0, Main.BACKGROUND_COLOR, this);
	}

	//----------------------------------------------------------------------

	// Mouse Dispatcher
	public class MouseDispatcher extends MouseAdapter implements MouseMotionListener {

		MainPanel mainPanel;

		MouseDispatcher(MainPanel mp) {
			mainPanel = mp;
		}

		//----------------------------------------------------------------------

		public void mouseMoved(MouseEvent e) { }
		
		//----------------------------------------------------------------------

		public void mousePressed(MouseEvent e) {
			Main.debugPrint(1, "mousePressed");

			if ((e.getModifiers() & InputEvent.BUTTON1_MASK) == 0) {
				return;
			}

			mainPanel.points = new LinkedList();

			Point p = e.getPoint();
			mainPanel.prevPoint.move(p.x, p.y);
		}

		//----------------------------------------------------------------------
		
		public void mouseDragged(MouseEvent e) { 
			if ((e.getModifiers() & InputEvent.BUTTON1_MASK) == 0) {
				return;
			}

			Dimension d = mainPanel.getSize();
			Point p = e.getPoint();

			points.add(p);
			
			// if the point is in the displayed area.
			if (0 <= p.x && p.x < d.width && 0 <= p.y && p.y < d.height) {

				Graphics graphics = mainPanel.getGraphics();

				Graphics2D g = (Graphics2D)graphics;
				
				g.setColor(Color.black);
				g.setStroke(Main.doubledStroke);
				g.draw(new Line2D.Double(prevPoint.x, prevPoint.y, p.x, p.y));
				
			}
			prevPoint.move(p.x, p.y);

		}       

		//----------------------------------------------------------------------
		
		public void mouseReleased(MouseEvent e) { 
			if ((e.getModifiers() & InputEvent.BUTTON1_MASK) == 0) {
				return;
			}

			mainPanel.prevPoint.move(-1, -1);

			LinkedList tokens = Recognition.tokenize(points);
			if (tokens == null) {
				return;
			}

			ListIterator it = tokens.listIterator();
			while (it.hasNext()) {
				Token t = (Token)it.next();
			}
			
			Recognition.recognize(points, tokens);
			
			mainPanel.repaint();
		}

		//----------------------------------------------------------------------
		
		public void mouseClicked(MouseEvent e) { 
			Point p = e.getPoint();
			if ((e.getModifiers() & InputEvent.BUTTON1_MASK) == 0) {
				Recognition.recognizeOnRightButtonClick(p);
			} else {
				Recognition.recognizeOnClick(p);
			}
			mainPanel.repaint();
		}
		
	}
}
