import java.awt.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;
import javax.sound.midi.*;

public class Play implements Runnable {

	static Synthesizer synthesizer = null;

	LinkedList formalized;
	int tempo;

	//----------------------------------------------------------------------

	Play(int _tempo, LinkedList _notes) {
		if (synthesizer == null) {
			try {
				synthesizer = MidiSystem.getSynthesizer();
			} catch (Exception e) {
				Main.failwith(e.getMessage());
			}
		}
		tempo = _tempo;
		formalized = formalize(_notes);
	}

	//----------------------------------------------------------------------
	
	// BeamedNote  Primitive Ѵ롣
	private LinkedList formalize(LinkedList notes) {
		LinkedList formalizedNotes = new LinkedList();
		ListIterator it = notes.listIterator();
		while (it.hasNext()) {
			Note note = (Note)it.next();
			if (note instanceof BeamedNote) {
				BeamedNote beamed = (BeamedNote)note;
				ListIterator lit = beamed.notes.listIterator();
				while (lit.hasNext()) {
					BasicNote bn = (BasicNote)lit.next();
					formalizedNotes.add(bn);
				}
			} else {
				formalizedNotes.add(note);
			}
		}

		return formalizedNotes;
	}

	//----------------------------------------------------------------------
	
	public void run() {
		try {
			ListIterator it = formalized.listIterator();
			
			synthesizer.open();
			Soundbank soundbank = synthesizer.getDefaultSoundbank();
			Instrument[] instruments = soundbank.getInstruments();
			synthesizer.loadInstrument(instruments[0]);
			
			MidiChannel channel = synthesizer.getChannels()[0];
			
			while (it.hasNext()) {
				Primitive prim = (Primitive)it.next();
				int ms = prim.getTime(tempo);
				System.out.println(ms);
				if (prim instanceof BasicNote) {
					BasicNote bn = (BasicNote)prim;
					int pitch = bn.getRealPitch();
					channel.noteOn(pitch, 127);
					Thread.currentThread().sleep(ms);
					channel.noteOff(pitch);
				} else if (prim instanceof Rest) {
					Thread.currentThread().sleep(ms);
				} else {
					Main.failwith("Play.play");
				}
			}
		} catch (Exception e) {
			Main.failwith(e.getMessage());
		}
	}
}
