import java.awt.*;
import javax.swing.*;
import javax.swing.table.*;
import java.awt.event.*;
import java.awt.geom.*;
import java.util.*;


class Token {
	int type;
	double length;
	
	Token(int t, double l) {
		type = t;
		length = l;
	}
}

//----------------------------------------------------------------------

public class Recognition {

	static final int E   = 1;  // East
	static final int ESE = 2;
	static final int SE  = 3;
	static final int SSE = 4;
	static final int S   = 5; // South
	static final int SSW = 6;
	static final int SW  = 7;
	static final int WSW = 8;
	static final int W   = 9;  // West
	static final int WNW = 10;
	static final int NW  = 11;
	static final int NNW = 12;
	static final int N   = 13;  // North
	static final int NNE = 14;
	static final int NE  = 15;	
	static final int ENE = 16;	

	static final Undo undo = Main.undo;

	//----------------------------------------------------------------------
   
	// ȡľ
	static private int toToken(Point p) {
		double d = Math.atan2(p.y, p.x) + 2 * Math.PI + Math.PI / 16;
		return (int)(d * 8 / Math.PI) % 16 + 1;
	}

	//----------------------------------------------------------------------
	
	// p 2ΥȤ
	static private double norm(Point p) {
		Main.debugPrint(0, "Norm : X = " + p.x + " Y = " + p.y);
		return Math.sqrt(p.x * p.x + p.y * p.y);
	}

	//----------------------------------------------------------------------
	
	//  a, b  threshold ƱѤȤߤʤ뤫ɤ
	static private boolean canUnify(int a, int b, int threshold) {
		int t = (a - b + 16) % 16;
		return (t <= threshold || (16 - threshold) <= t);
	}

	//----------------------------------------------------------------------
	
	// 줿 tokenize 롣
	static LinkedList tokenize(LinkedList points) {

		if (points == null || points.isEmpty()) {
			return null;
		}

		Point difference   = new Point(0, 0);
		LinkedList tokens  = new LinkedList();
		ListIterator it   = points.listIterator();
		Point prevPoint   = (Point)it.next();
		Point currentPoint;
		Token token = null;

		while (it.hasNext()) {
			currentPoint = (Point)it.next();
			difference.setLocation(currentPoint.x - prevPoint.x,
								   currentPoint.y - prevPoint.y);
			int t = toToken(difference);

			// ޤᤤϡ̵뤹롣
			if (norm(difference) < 5.0) {
				if (token == null || token.type != t) {
					continue;
				} else {
					token.length += norm(difference);
				}
			} else {
				if (token == null) {
					token = new Token(t, norm(difference));
				} else if (token.type == t) {
					token.length += norm(difference);
				} else {
					tokens.add(token);
					token = new Token(t, norm(difference));
				}
				
			}

			prevPoint = currentPoint;

		}
		
		if (token == null) {
			return tokens;
		} else {
			tokens.add(token);
			return tokens;
		}

	}

	//----------------------------------------------------------------------
	
	static private void recognizeOnSpace(Field field, Point start, LinkedList tokens) {

		if (tokens == null || start == null) {
			return;
		}

		// the following code is state-based.
		// you must think it too ugly to read. (Even I think so!)
		// However... Is there another (more beautiful) coding?

		ListIterator it = tokens.listIterator();
		boolean onLine = false;
		int state = 0;

		while (it.hasNext()) {
			Token token = (Token)it.next();
			int t = token.type;
			switch (state) {
			case 0:
				if (canUnify(t, E, 2)) {
					onLine = true;
					state = 1;
				} else if (canUnify(t, N, 2)) {
					state = 2;
				} else if (canUnify(t, S, 2)) {
					state = 7;
				} else if (canUnify(t, W, 2)) {
					state = 6;
				} else {
					state = -1;
				}
				break;
			case 1:
				if (canUnify(t, E, 2)) {
					state = 1;
				} else if (canUnify(t, N, 2)) {
					state = 2;
				} else {
					state = -1;
				}
				break;
			case 2:
				if (canUnify(t, N, 2)) {
					state = 2;
				} else if (canUnify(t, SE, 3)) {
					state = 3;
				} else {
					state = -1;
				}
				break;
			case 3:
				if (canUnify(t, SE, 3)) {
					state = 3;
				} else if (canUnify(t, N, 3)) {
					state = 4;
				} else {
					state = -1;
				}
				break;
			case 4:
				if (canUnify(t, N, 3)) {
					state = 4;
				} else if (canUnify(t, SE, 3)) {
					state = 5;
				} else {
					state = -1;
				}
				break;
			case 5:
				if (canUnify(t, SE, 3)) {
					state = 5;
				} else {
					state = -1;
				}
				break;
			case 6:
				if (canUnify(t, W, 2)) {
					state = 6;
				} else {
					state = -1;
				}
				break;
			case 7:
				if (canUnify(t, S, 2)) {
					state = 7;
				} else if (canUnify(t, NE, 2)) {
					state = 8;
				} else {
					state = -1;
				}
				break;
			case 8:
				if (canUnify(t, NE, 2)) {
					state = 8;
				} else if (canUnify(t, S, 3)) {
					state = 9;
				} else {
					state = -1;
				}
				break;
			case 9:
				if (canUnify(t, S, 3)) {
					state = 9;
				} else if (canUnify(t, NE, 3)) {
					state = 10;
				} else {
					state = -1;
				}
				break;
			case 10:
				if (canUnify(t, NE, 3)) {
					state = 10;
				} else {
					state = -1;
				}
			default:
				break;
			}
			
		}
			
		// 1  whole note
		// 2  quarter note
		// 3  eighth note
		// 4  sixteenth note
		// 5  thirtysecond note
		// 6  whole rest
		// 7  quarter rest
		// 8  eighth rest
		// 9  sixteenth rest
		// 10 thirtysecond rest

		if (state <= 0) {
			return;
		}

		Note note = null;

		if (state == 1) {
			int s = Main.score.getRelativeYOfField(start.y);
			int pitch = field.getPitchFromRelativeY(s);
			note = new BasicNote(field, pitch, BasicNote.NOTE1, 0, start.x);
		} else if (6 <= state && state <= 10) {
			int type = 0;
			switch (state) {
			case 6:
				type = Rest.REST1;
				break;
			case 7:
				type = Rest.REST4;
				break;
			case 8:
				type = Rest.REST8;
				break;
			case 9:
				type = Rest.REST16;
				break;
			case 10:
				type = Rest.REST32;
				break;
			} 

			note = new Rest(field, type, 0, start.x);
		} else {
			int type = 0;
			switch (state) {
			case 2:
				type = BasicNote.NOTE4;
				break;
			case 3:
				type = BasicNote.NOTE8;
				break;
			case 4:
				type = BasicNote.NOTE16;
				break;
			case 5:
				type = BasicNote.NOTE32;
				break;
			}
			int s = Main.score.getRelativeYOfField(start.y);
			int pitch = field.getPitchFromRelativeY(s, onLine);
			
			note = new BasicNote(field, pitch, type, 0, start.x);
		}

		field.addNoteWithUndo(note, undo);
	}

	//----------------------------------------------------------------------

	static private boolean canDelete(LinkedList tokens) {
		if (tokens == null || tokens.isEmpty()) {
			return false;
		}

		int args = 0;
		ListIterator it = tokens.listIterator();
		int prevDir = ((Token)it.next()).type;
		while (it.hasNext()) {
			Token token = (Token)it.next();
			int   t     = token.type;
			args += 8 - Math.abs((Math.abs(t - prevDir)) - 8); // г (15 ñ)
		}

		return (args >= 20);
	}

	//----------------------------------------------------------------------

	static private void recognizeDeletion(Field field, LinkedList tokens, Note note) {
		if (canDelete(tokens)) {
			field.removeNoteWithUndo(note, undo);
		}
	}

	static private void recognizeDeletion(Field field, LinkedList tokens, LinkedList notes) {
		if (canDelete(tokens)) {
			field.removeNotesWithUndo(notes, undo);
		}
	}

	//----------------------------------------------------------------------

	static private void recognizeOnNotehead(Field field, Point startPoint, LinkedList tokens, BasicNote note) {
		
		ListIterator it = tokens.listIterator();
		int state = 0;
		double lengthNE = 0.0;
		double lengthE  = 0.0;
		double lengthSE = 0.0;

		while (it.hasNext()) {
			Token token = (Token)it.next();
			int   t     = token.type;
			switch (state) {
			case 0:
				if (canUnify(t, N, 1)) {
					state = 1;
					break;
				} else if (canUnify(t, E, 3)){
					state = 2;
					switch (t) {
					case NNE:
						lengthNE += token.length / 2;
						break;
					case NE:
						lengthNE += token.length;
						break;
					case ENE:
						lengthNE += token.length / 2;
						lengthE  += token.length / 2;
						break;
					case E:
						lengthE  += token.length;
						break;
					case ESE:
						lengthE  += token.length / 2;
						lengthSE += token.length / 2;
						break;
					case SE:
						lengthSE += token.length;
						break;
					case SSE:
						lengthSE += token.length / 2;
						break;
					default:
						state = -1;
						break;
					}
				} else {
					state = -1;
				}
				break;
			case 1:
				// 
				if (canUnify(t, N, 1)) {
					break;
				} else {
					state = -1;
				}
			case 2:
				if (canUnify(t, E, 3)){
					switch (t) {
					case NNE:
						lengthNE += token.length / 2;
						break;
					case NE:
						lengthNE += token.length;
						break;
					case ENE:
						lengthNE += token.length / 2;
						lengthE  += token.length / 2;
						break;
					case E:
						lengthE  += token.length;
						break;
					case ESE:
						lengthE  += token.length / 2;
						lengthSE += token.length / 2;
						break;
					case SE:
						lengthSE += token.length;
						break;
					case SSE:
						lengthSE += token.length / 2;
						break;
					default:
						state = -1;
						break;
					}
				} else {
					state = -1;
					break;
				}
			default:
				break;
			}
		}

		if (state == 1) {
			if (note.type == BasicNote.NOTE1) {
				note.halveNoteWithUndo(undo);
				return;
			}
		} else if (state == 2) {
			// Hmm...
			if (lengthNE > lengthE) {
				if (lengthNE > lengthSE) {
					// lengthNE
					if (lengthSE <= 0.0) {
						note.modifyNoteWithUndo(BasicNote.SHARP, undo);
						return;
					} else {
						return;
					}
				} else {
					// lengthSE
					if (lengthNE <= 0.0) {
						note.modifyNoteWithUndo(BasicNote.FLAT, undo);
					}
					return;
				}
			} else {
				if (lengthE > lengthSE) {
					// lengthE
					note.modifyNoteWithUndo(BasicNote.NATURAL, undo);
					return;
				} else {
					// lengthSE
					if (lengthNE <= 0.0) {
						note.modifyNoteWithUndo(BasicNote.FLAT, undo);
					}
					return;
				}
			}
		} else {
			recognizeDeletion(field, tokens, note);
		}
	}

	//----------------------------------------------------------------------

	static private void recognizeSingleNoteCrossed(Field field, Note n, LinkedList tokens) {
		// In this case, note should be instance of BasicNote or Rest.
		
		ListIterator it = tokens.listIterator();
		int state = 0;

		while (it.hasNext()) {
			Token token = (Token)it.next();
			int t = token.type;
			switch (state) {
			case 0:
				if (canUnify(t, NE, 2)) {
					state = 1;
				} else if (canUnify(t, NW, 2)) {
					state = 2;
				} else {
					state = -1;
				}
				break;
			case 1:
				if (canUnify(t, NE, 2)) {
					break;
				} else {
					state = -1;
				}
				break;
			case 2:
				if (canUnify(t, NW, 2)) {
					break;
				} else {
					state = -1;
				}
				break;
			default:
				break;
			}
		}

		if (state == 1) {
			n.halveNoteWithUndo(undo);
		} else if (state == 2) {
			n.doubleNoteWithUndo(undo);
		} else {
			recognizeDeletion(field, tokens, n);
		}

	}

	//----------------------------------------------------------------------

	static private void recognizeBasicNotesCrossed(Field field, LinkedList notes, LinkedList tokens) {
		int state = 0;
		ListIterator it = tokens.listIterator();
		while (it.hasNext()) {
			Token token = (Token)it.next();
			int t = token.type;
			switch (state) {
			case 0:
				if (canUnify(t, E, 2)) {
					state = 1;
				} else if (canUnify(t, W, 2)) {
					state = 2;
				} else {
					state = -1;
				}
				break;
			case 1:
				if (canUnify(t, E, 2)) {
					break;
				} else {
					state = -1;
				}
				break;
			case 2:
				if (canUnify(t, W, 2)) {
					break;
				} else {
					state = -1;
				}
				break;
			default:
				break;
			}
		}

		if (state == 1 || state == 2) {
			field.beamNoteWithUndo(notes, undo);
		} else {
			recognizeDeletion(field, tokens, notes);
		}

	}

	//----------------------------------------------------------------------

	static void recognizeSingleBeamedNoteCrossed(Field field, BeamedNote beamed, 
												 LinkedList points, LinkedList tokens) {
		
		// beamed ΤĤ˥Ƥ뤫ޤ롣
		LinkedList crossedPrimitives = getCrossedPrimitives(field, points, tokens);
		if (crossedPrimitives.isEmpty()) {
			return;
		} else if (crossedPrimitives.size() == 1) { // Ĥ
			Primitive prim = (Primitive)crossedPrimitives.getFirst();
			if (prim instanceof BasicNote) {
				Point p = (Point)points.getFirst();
				int relativeY  = Main.score.getRelativeYOfField(p.y);
				BasicNote bn = (BasicNote)prim;
				if (bn.isOnNotehead(p.x, relativeY)) {
					recognizeOnNotehead(field, p, tokens, bn);
				} else {
					recognizeDeletion(field, tokens, bn);
				}
			} else {
				recognizeDeletion(field, tokens, prim);
			}
			return;
		} else if (crossedPrimitives.size() == beamed.notes.size()) { // 
			int state = 0;
			ListIterator it = tokens.listIterator();
			while (it.hasNext()) {
				Token token = (Token)it.next();
				int t = token.type;
				switch (state) {
				case 0:
					if (canUnify(t, E, 2)) {
						state = 1;
					} else if (canUnify(t, W, 2)) {
						state = 2;
					} else {
						state = -1;
					}
					break;
				case 1:
					if (canUnify(t, E, 2)) {

					} else {
						state = -1;
					}
					break;
				case 2:
					if (canUnify(t, W, 2)) {
						
					} else {
						state = -1;
					}
					break;
				default:
					break;
				}
			}

			if (state == 1) {
				beamed.halveNoteWithUndo(undo);
			} else if (state == 2) {
				beamed.doubleNoteWithUndo(undo);
			} else {
				recognizeDeletion(field, tokens, beamed);
			}

		} else {
			recognizeDeletion(field, tokens, beamed);
		}

		return;
	}

	//----------------------------------------------------------------------

	static public void recognizeOnClick(Point p) {
		// Field field = Main.score.getFieldFromY(p.y);

		Primitive prim = Main.score.getPrimitiveFromXY(p.x, p.y);

		if (prim == null) {
			return;
		}

		if (prim instanceof BasicNote) {
			BasicNote bn = (BasicNote)prim;
			int relativeY = Main.score.getRelativeYOfField(p.y);
			if (bn.isOnNotehead(p.x, relativeY)) {
				bn.addDotWithUndo(undo);
			} 
		} else if (prim instanceof Rest) {
			Rest rest = (Rest)prim;
			rest.addDotWithUndo(undo);
		} else {
			Main.failwith("recognizeOnClick : wrong type?");
		}

	}

	//----------------------------------------------------------------------

	static public void recognizeOnRightButtonClick(Point p) {

		Primitive prim = Main.score.getPrimitiveFromXY(p.x, p.y);
		if (prim == null) {
			return;
		}

		prim.removeAllModificationWithUndo(undo);

	}

	//----------------------------------------------------------------------

	static private LinkedList getCrossedNotes(Field field, LinkedList points, LinkedList tokens) {

		// Ƥ벻ƽʣ󤹤롣
		LinkedList crossedNotes = new LinkedList();
		ListIterator it = points.listIterator();
		while (it.hasNext()) {
			Point p = (Point)it.next();
			Note  n = field.getNoteFromX(p.x);
			// ˻äƤϡʤ
			if (n == null || crossedNotes.contains(n)) {
				continue;
			} else if (crossedNotes.isEmpty()) {
				crossedNotes.add(n);
			} else { // ɲä뤳ȤɬפʤȤ
				ListIterator lit = crossedNotes.listIterator();
				int index = 0;
				while (lit.hasNext()) {
					Note note = (Note)lit.next();
					if (n.horizonPosition < note.horizonPosition) {
						break;
					}
					++index;
				}
				crossedNotes.add(index, n);
			}
		}

		return crossedNotes;
	}

	//----------------------------------------------------------------------

	static private LinkedList getCrossedPrimitives(Field field, LinkedList points, LinkedList tokens) {

		// Ƥ Primitive ƽʣ󤹤롣
		LinkedList crossedPrimitives = new LinkedList();
		ListIterator it = points.listIterator();
		while (it.hasNext()) {
			Point p = (Point)it.next();
			Note  n = field.getPrimitiveFromX(p.x);
			// ˻äƤϡʤ
			if (n == null || crossedPrimitives.contains(n)) {
				continue;
			} else if (crossedPrimitives.isEmpty()) {
				crossedPrimitives.add(n);
			} else { // ɲä뤳ȤɬפʤȤ
				ListIterator lit = crossedPrimitives.listIterator();
				int index = 0;
				while (lit.hasNext()) {
					Primitive note = (Primitive)lit.next();
					if (n.horizonPosition < note.horizonPosition) {
						break;
					}
					++index;
				}
				crossedPrimitives.add(index, n);
			}
		}

		return crossedPrimitives;
	}

	//----------------------------------------------------------------------
	
	static void recognize(LinkedList points, LinkedList tokens) {

		if (points.isEmpty()) {
			return;
		}

		Field field = Main.score.getFieldFromY(((Point)points.getFirst()).y);
		if (field == null) {
			return;
		}

		// Ƥ Note 롣
		LinkedList crossedNotes = getCrossedNotes(field, points, tokens);

		// Ƥ벻򸫤롣
		if (crossedNotes.isEmpty()) {
			recognizeOnSpace(field, (Point)points.getFirst(), tokens);
			return;
		} 

		if (crossedNotes.size() == 1) {
			Note n = (Note)crossedNotes.getFirst();
			Point p = (Point)points.getFirst();
			if (n instanceof BasicNote) {
				BasicNote note = (BasicNote)n;
				int relativeY  = Main.score.getRelativeYOfField(p.y);
				if (note.isOnNotehead(p.x, relativeY)) {
					recognizeOnNotehead(field, p, tokens, note);
				} else {
					recognizeSingleNoteCrossed(field, note, tokens);
				}
				return;
			} else if (n instanceof Rest) {
				Rest note = (Rest)n;
				recognizeSingleNoteCrossed(field, note, tokens);
				return;
			} else if (n instanceof BeamedNote) {
				BeamedNote b = (BeamedNote)n;
				recognizeSingleBeamedNoteCrossed(field, b, points, tokens);
				return;
			} else {
				Main.failwith("What happened?");
			}

		}

		// ʣ Note ˸٤äƤ
		boolean isAllBasicNote = true;

		ListIterator iter = crossedNotes.listIterator();
		while (iter.hasNext()) {
			Note n = (Note)iter.next();
			System.out.println(n.toString());
			if (! (n instanceof BasicNote)) {
				isAllBasicNote = false;
			}
		}

		if (isAllBasicNote) {
			System.out.println("All Selected Notes are BasicNote.");
			recognizeBasicNotesCrossed(field, crossedNotes, tokens);
		} else {
			recognizeDeletion(field, tokens, crossedNotes);
		}

	} 

}
